-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, technology_lib, tech_flash_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY mms_remu IS
  GENERIC MAP (
    g_technology : NATURAL := c_tech_select_default
  );
  PORT (
    mm_rst          : IN    STD_LOGIC;
    mm_clk          : IN    STD_LOGIC;

    epcs_clk        : IN    STD_LOGIC;

    -- MM registers
    remu_mosi       : IN    t_mem_mosi := c_mem_mosi_rst;
    remu_miso       : OUT   t_mem_miso  
  );
END mms_remu;


ARCHITECTURE str OF mms_remu IS

  CONSTANT c_param_w       : NATURAL := 3;
  CONSTANT c_data_w        : NATURAL := 24;
 
  SIGNAL epcs_rst          : STD_LOGIC;

  SIGNAL remu_busy         : STD_LOGIC;
  SIGNAL remu_data_out     : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0);

  SIGNAL fall_remu_reconfigure  : STD_LOGIC;
  SIGNAL fall_remu_read_param   : STD_LOGIC;
  SIGNAL fall_remu_write_param  : STD_LOGIC;
  SIGNAL fall_remu_param        : STD_LOGIC_VECTOR(c_param_w-1 DOWNTO 0);
  SIGNAL fall_remu_data_in      : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0);

  SIGNAL nxt_fall_remu_reconfigure  : STD_LOGIC;
  SIGNAL nxt_fall_remu_read_param   : STD_LOGIC;
  SIGNAL nxt_fall_remu_write_param  : STD_LOGIC;
  SIGNAL nxt_fall_remu_param        : STD_LOGIC_VECTOR(c_param_w-1 DOWNTO 0);
  SIGNAL nxt_fall_remu_data_in      : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0);

BEGIN

  u_remu: ENTITY tech_flash_lib.tech_flash_remote_update
  GENERIC MAP (
    g_technology    => g_technology
  )
  PORT MAP (        
    clock         => epcs_clk, 
    param         => fall_remu_param,  
    read_param    => fall_remu_read_param,   
    reconfig      => fall_remu_reconfigure,
    reset         => epcs_rst,
    reset_timer   => '0',
    busy          => remu_busy,
    data_out      => remu_data_out,
    write_param   => fall_remu_write_param,
    data_in       => fall_remu_data_in
   );

  u_remu_reg: ENTITY work.remu_reg                   
  PORT MAP (       
    mm_rst            => mm_rst, 
    mm_clk            => mm_clk,

    epcs_rst          => epcs_rst, 
    epcs_clk          => epcs_clk,

    sla_in            => remu_mosi,
    sla_out           => remu_miso,

    reconfigure       => nxt_fall_remu_reconfigure,
    read_param        => nxt_fall_remu_read_param,
    param             => nxt_fall_remu_param,
    busy              => remu_busy,

    data_out          => remu_data_out,
    write_param       => nxt_fall_remu_write_param,
    data_in           => nxt_fall_remu_data_in
  );

  -- The Remote Upgrade megafunction uses an ASMI_PARALLEL instance internally and therefore requires
  -- the same clocking scheme as the EPCS module:

  -- The ASMI_PARALLEL requires that data is set up on the falling edge. The following rising
  -- edge of the ASMI_PARALLEL's 20MHz clkin clocks in the data. The data is required to be held until
  -- the next falling edge:
  --  __ __ __ __ __ __
  -- _||_||_||_||_||_||_
  --         ____
  -- ________|  |_______
  --
  -- All used components expect rising edge clocked signals, so we're reclocking the relevant signals
  -- with falling-edge clocked registers: 
  p_epcs_clk: PROCESS(epcs_clk, epcs_rst)
  BEGIN
    IF epcs_rst='1' THEN      
      fall_remu_reconfigure <= '0';
      fall_remu_read_param  <= '0';
      fall_remu_write_param <= '0';
      fall_remu_param       <= (OTHERS=>'0');
      fall_remu_data_in     <= (OTHERS=>'0');
    ELSIF falling_edge(epcs_clk) THEN
      fall_remu_reconfigure <= nxt_fall_remu_reconfigure;
      fall_remu_read_param  <= nxt_fall_remu_read_param;
      fall_remu_write_param <= nxt_fall_remu_write_param;
      fall_remu_param       <= nxt_fall_remu_param;
      fall_remu_data_in     <= nxt_fall_remu_data_in;
    END IF;
  END PROCESS;

  u_common_areset: ENTITY common_lib.common_areset                       
  PORT MAP (        
    in_rst             => mm_rst, 
    clk                => epcs_clk,        
    out_rst            => epcs_rst  
  );
  
END str;

