###############################################################################
#
# Copyright (C) 2014
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

"""Common class for accessing a Python dictionary in one or more files

   The contents of the dictionary file consist of a series of key - value
   pairs. These key - value pairs are read from the file and kept in a
   single dictionary of keys and values.

   The format of the dictionary file is similar to that of an ini file but
   without the sections. For ini files Python has the ConfigParser package.
   Effectively the dictionary file only contains one section. It is not
   obvious how to deal with sections if only one dictionary needs to be
   stored flat. Defining a single section, is misleading because it suggests
   that more sections are supported. Alternatively prepending a dummy
   section when the file is read is also a bit awkward. In fact the parsing
   by read_dict_file() is relatively simple and the other methods would also
   be needed if the ConfigParser would be used. Therefore the Python
   ConfigParser package for ini files is not used.

   The key and value string in the dictionary file are separated by '='. Hence
   the '=' character can not be used in keys. The '=' can be used in values,
   because subsequent '=' on the same line are part of the value.
   Every key must start on a new line. The value string can extend over one
   or multiple lines. Comment lines are supported by preceding it with a '#'
   and will get stripped.

     # This is a comment section
     # a key starts on a new line and extends until the '='
     # a key and its values are separated by '='

     key=string
     key =string
     key = string

     # a key with multiple values in its string
     key = value value value

     # how the values are separated depends on the dictionary user
     key = value, value, value
     key = value value

     # a key with many values can have its string extend on multiple lines,
     # the newline is replaced by a ' ' and any indent is removed
     key =
     value
     value
     value

     # empty lines and spaces are allowed
     key = value

   Whether the key is a valid key depends on the dict user that interprets
   the dictionary.

   It can be useful to be able to refer to the dicts based on a certain key
   value, e.g. a key value that gives the name of the dict. The methods
   get_key_values() and get_dicts() provide a means to easily link a dict
   to its name and its name to the dict. The argument and the return can be
   a list or a single value. Default the init list of self.dicts is used as
   argument.

   'fileName'     = the name of the dictionary file including extension
   'filePath'     = the full directory name where the dictionary file is
                    stored
   'filePathName' = filePath/fileName

"""

import common as cm
import sys
import os
import os.path
import collections

class CommonDictFile:

    def __init__(self, rootDir, fileName='dict.txt'):
        """Store the dictionaries from all fileName files in rootDir."""
        self.CDF_COMMENT = '#'
        self.CDF_SEPARATOR = '='
        self.rootDir = rootDir
        self.fileName = fileName                            # all dictionary files have the same fileName
        self.filePaths = self.find_all_dict_file_paths()    # list of all directory paths of dictionary files that are available in the rootDir tree
        if len(self.filePaths)==0:
            sys.exit('Error : No %s file found in %s directory tree.' % (fileName, rootDir))
        self.filePathNames = []                             # list of all directory paths + fileName of the available dictionary files
        for path in self.filePaths:
            self.filePathNames.append(os.path.join(path, self.fileName))
        self.dicts = self.read_all_dict_files()             # list of dictionaries that are read from the available dictionary files
        self.nof_dicts = len(self.dicts)                    # number of dictionaries = number of dictorary files

    def find_all_dict_file_paths(self, rootDir=None):
        """Recursively search the rootDir tree to find the paths to all fileName files."""
        if rootDir==None: rootDir=self.rootDir
        paths = []
        for root, _, files in os.walk(rootDir):
            if self.fileName in files:
                 paths.append(root)
        return paths

    def read_all_dict_files(self, filePathNames=None):
        """Read the dictionary information from all files that were found in the rootDir tree."""
        if filePathNames==None: filePathNames=self.filePathNames
        read_dicts = []
        for fp in self.filePathNames:
            read_dicts.append(self.read_dict_file(fp))
        return read_dicts

    def read_dict_file(self, filePathName=None, valueSeparator=' '):
        """Read the dictionary information the filePathName file."""
        if filePathName==None: filePathName=self.filePathNames[0]
        file_dict = collections.OrderedDict()
        with open(filePathName, 'r') as fp:
            lineNr=0
            key = ''
            value = ''
            for line in fp:
                if line.find(self.CDF_COMMENT)==-1:          # strip lines with comment
                    key_end = line.find(self.CDF_SEPARATOR)  # find key
                    if key_end>=0:
                        key = line[0:key_end].strip()        # new key
                        value = line[key_end+1:].strip()     # new value
                    else:
                        value += valueSeparator              # replace newline by valueSeparator
                        value += line.strip()                # append value
                    if key!='':
                        file_dict[key] = value.strip()       # update key in dict
                lineNr += 1
        return file_dict

    def write_dict_file(self, dicts, filePathNames, keySeparator=None):
        if keySeparator==None: keySeparator=self.CDF_SEPARATOR
        for fpn, the_dict in zip(cm.listify(filePathNames), cm.listify(dicts)):
            with open(fpn, 'w') as fp:
                for key in the_dict:
                    fp.write('%s%s%s\n' % (key, keySeparator, the_dict[key]))

    def append_key_to_dict_file(self, filePathName, key, values):
        with open(filePathName, 'a') as fp:
            if len(cm.listify(values))==1:
                fp.write('%s = %s' % (key, values))
            else:
                fp.write('%s = \n' % key)
                for v in cm.listify(values):
                    fp.write('%s\n' % v)

    def get_filePath(self, the_dict):
        return self.filePaths[self.dicts.index(the_dict)]

    def get_filePathName(self, the_dict):
        return self.filePathNames[self.dicts.index(the_dict)]

    def get_key_values(self, key, dicts=None):
        """Get the value of a key in the dicts or None in case the key does not exist."""
        if dicts==None: dicts=self.dicts
        key_values = []
        for fd in cm.listify(dicts):
            if key in fd:
                key_values.append(fd[key])
            else:
                key_values.append(None)
        return cm.unlistify(key_values)

    def get_dicts(self, key, values, dicts=None):
        """Get all dictionaries in dicts that contain a key with a value specified in values."""
        if dicts==None:
            dicts=self.dicts
        the_dicts = []
        for fd in cm.listify(dicts):
            if fd not in the_dicts:
                if key in fd:
                    if fd[key] in cm.listify(values):
                       the_dicts.append(fd)
        return cm.unlistify(the_dicts)


if __name__ == '__main__':
    tmpFileName = 'tmp_dict.txt'
    tmpDirName = 'tmp_dict'

    # Create some example file
    with open(tmpFileName, 'w') as f:
        f.write('# Example dictionary file 0\n')
        f.write('src=\n')
        f.write('z.vhd\n')
        f.write('\n')
        f.write('u.vhd\n')
        f.write('tb = \'c.vhd\n')
        f.write('x.vhd\'\n')
        f.write('abc.vhd\n')
        f.write('syn=\n')
        f.write(' = \n')
        f.write('              \n')
        f.write('sim= d, u\n')
        f.write('test= d u\n')
        f.write('src_files = \n')
        f.write('     x.vhd \n')
        f.write('    y.vhd \n')
        f.write('equal= with = in value\n')
        f.write('\n')

    # and created another example file in a sub directory
    os.mkdir(tmpDirName)
    tmpFilePathName = os.path.join(tmpDirName, tmpFileName)
    with open(tmpFilePathName, 'w') as f:
        f.write('# Example dictionary file 1\n')
        f.write('   # Strip indent comment\n')
        f.write('skip = will get skipped due to # this indent comment\n')
        f.write('src = x.vhd y.vhd\n')

    # Read the dictionary from the example files
    cdf = CommonDictFile(rootDir='./', fileName=tmpFileName)

    print '#'
    print '# Test for CommonDictFile.py'
    print '#'
    print 'rootDir       = %s' % cdf.rootDir
    print 'nof_dicts     = %d' % cdf.nof_dicts
    print 'fileName      = %s' % cdf.fileName
    print 'filePaths     = %s' % cdf.filePaths
    print 'filePathNames = %s' % cdf.filePathNames
    print ''
    for i, p in enumerate(cdf.filePaths):
        print os.path.join(p, cdf.fileName)
        d = cdf.dicts[i]
        for k,v in d.iteritems():
            print k, '=', v
        print ''

    # Write dict file
    cdf.write_dict_file(cdf.dicts[0], 'tmp_dict.out')

    # Remove the example files
    os.remove(tmpFileName)
    os.remove(tmpFilePathName)
    os.rmdir(tmpDirName)