###############################################################################
#
# Copyright (C) 2014
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

"""HDL configuration for building simulation and synthesis targets.

   There should be one hdltool.cfg file somewhere in the toolRootDir and at
   least one hdllib.cfg file somewhere in the libRootDir. Every HDL library
   that is in the libRootDir can be found if it has a hdllib.cfg file.
   Together the hdltool.cfg and hdllib.cfg files contain all the keys and
   values that are sufficient to be able to build the targets for the HDL
   library. The possible targets are:

   - compile to created the library binaries for simulation
   - synthesize to created an image that can be loaded ion the FPGA
   - verify VHDL test benches in simulation
   - verify Python test cases via the MM control interface in simulation
   - validate Python test cases on hardware via the MM control interface

   The contents of the cfg files consist of a series of key - value pairs
   that are read into a dictionary as defined in common_dict_file.py. Whether
   the key is a valid key depends on the application that interprets the
   dictionary.

   The methods can have the library dictionary or the library name as
   argument. The default arguments are the self.libs.dicts and the
   corresponding self.lib_names. The argument can be a list or a single value.
   Similar the return can be a list or a single value, because a list of one
   element is unlistified.

"""

import common as cm
import common_dict_file
import sys
import os
import os.path

class HdlConfig:

    def __init__(self, libRootDir, toolRootDir, libFileName='hdllib.cfg', toolFileName='hdltool.cfg'):
        """Get tool dictionary info from toolRootDir and all HDL library dictionary info from libRootDir."""
        self.libRootDir = libRootDir
        self.toolRootDir = toolRootDir
        # HDL tool config file
        self.tool = common_dict_file.CommonDictFile(toolRootDir, toolFileName)      # tool dict file
        if self.tool.nof_dicts==0: sys.exit('Error : No HDL tool config file found')
        if self.tool.nof_dicts >1: sys.exit('Error : Multiple HDL tool config files found')
        # HDL library config files
        self.libs = common_dict_file.CommonDictFile(libRootDir, libFileName)        # library dict files
        if self.libs.nof_dicts==0: sys.exit('Error : No HDL library config file found')
        self.lib_names = self.libs.get_key_values('hdl_lib_name')
        self.MODELSIM_MPF_NAME = 'hdllib.mpf'

    def derive_all_use_libs(self, lib_name):
        """Derive a list of all HDL libraries that the specified HDL lib_name library depends on."""
        if lib_name in self.lib_names:
            all_use_libs = [lib_name]
            lib_dict = self.libs.dicts[self.lib_names.index(lib_name)]
            if 'hdl_lib_uses' in lib_dict:
                use_libs = lib_dict['hdl_lib_uses'].split()
                for use_lib in use_libs:
                    all_use_libs.append(use_lib)
                    all_use_libs += self.derive_all_use_libs(use_lib)
            return cm.unique(all_use_libs)
        else:
            sys.exit('Error : Unknown HDL library name')

    def derive_lib_order(self, lib_names=None):
        """Derive the dependency order for all HDL libraries in the specified list of lib_names."""
        if lib_names==None: lib_names=self.lib_names
        lib_dicts = self.libs.get_dicts('hdl_lib_name', lib_names)
        lib_order = []
        for lib_dict in cm.listify(lib_dicts):
            lib_name = lib_dict['hdl_lib_name']
            if not lib_name in lib_order:
                lib_order.append(lib_name)                                        # append this lib
            if 'hdl_lib_uses' in lib_dict:
                use_libs = lib_dict['hdl_lib_uses'].split()
                for use_lib in use_libs:
                    if not use_lib in lib_order:
                        lib_order.insert(lib_order.index(lib_name), use_lib)      # insert used lib just before this lib
                    else:
                        if lib_order.index(use_lib) > lib_order.index(lib_name):
                            lib_order.remove(use_lib)
                            lib_order.insert(lib_order.index(lib_name), use_lib)  # move used lib to just before this lib
        return lib_order

    def get_lib_build_dirs(self, build_type, build_subdir=None, lib_dicts=None):
        """Get the tool build directory for all HDL libraries in the specified list of lib_dicts.
        
        The build_type can be:
            'sim'   uses the 'build_sim_dir'   key in the lib_dict,
            'synth' uses the 'build_synth_dir' key in the lib_dict
            
        If the build_subdir is not specfied then it defaults to the tool name key value.
        
        If the build dir key in HDL library config file is '' then
            treat the build dir key as local directory in the HDL library directory.
        else
            treat the build dir key as central directory.
        """
        if lib_dicts==None: lib_dicts=self.libs.dicts
        if build_type=='sim':
            build_dir_key = 'build_sim_dir'
            tool_name_key = 'sim_tool_name'
        elif build_type=='synth':
            build_dir_key = 'build_synth_dir'
            tool_name_key = 'synth_tool_name'
        else:
            sys.exit('Error : Unknown build type')
        if build_subdir==None: build_subdir = self.tool.get_key_values(tool_name_key)
        build_dirs = []
        for lib_dict in cm.listify(lib_dicts):
            build_dir = os.path.expandvars(self.libs.get_key_values(build_dir_key, lib_dict))
            if build_dir=='':
                lib_path = self.libs.get_filePath(lib_dict)
                build_dirs.append(os.path.join(lib_path, build_subdir))             # local build directory in HDL library directory
            else:
                lib_name = lib_dict['hdl_lib_name']
                build_dirs.append(os.path.join(build_dir, build_subdir, lib_name))  # central build cdirectory
        return cm.unlistify(build_dirs)

    def create_lib_order_files(self, lib_names=None):
        """Create the compile order file for all HDL libraries in the specified list of lib_names."""
        if lib_names==None: lib_names=self.lib_names
        lib_dicts = self.libs.get_dicts('hdl_lib_name', lib_names)
        for lib_dict in cm.listify(lib_dicts):
            lib_name = lib_dict['hdl_lib_name']
            use_libs = self.derive_all_use_libs(lib_name)
            lib_order = self.derive_lib_order(use_libs)
            file_name = lib_name + '_lib_order.txt'
            file_path = self.get_lib_build_dirs('sim', lib_dicts=lib_dict)
            cm.mkdir(file_path)
            filePathName = os.path.join(file_path, file_name)
            with open(filePathName, 'w') as fp:
                for lib in lib_order:
                    fp.write('%s ' % lib)

if __name__ == '__main__':
    # Read the dictionary info from all HDL tool and library configuration files in the current directory and the sub directories
    libRootDir = 'RADIOHDL'
    #libRootDir = 'UNB'
    hdl = HdlConfig(libRootDir=os.environ[libRootDir], toolRootDir=os.path.expandvars('$RADIOHDL/tools'), libFileName='hdllib.cfg', toolFileName='hdltool.cfg')

    print '#'
    print '# HdlConfig:'
    print '#'
    for i, p in enumerate(hdl.libs.filePathNames):
        print i, p
        d = hdl.libs.dicts[i]
        for k,v in d.iteritems():
            print k, '=', v
        print ''

    print ''
    print 'tool = ', hdl.tool.filePathNames[0]
    
    print ''
    print 'lib paths'
    for p in hdl.libs.filePaths:
        print '    ', p
        
    print ''
    print 'lib paths names'
    for p in hdl.libs.filePathNames:
        print '    ', p
        
    print ''
    print 'lib_names        = ', hdl.lib_names
    print 'derive_lib_order : ', hdl.derive_lib_order()

    print ''
    print 'get_lib_build_dirs for simulation:'
    for sim_dir in hdl.get_lib_build_dirs('sim'):
        print '    ', sim_dir
    
    print ''    
    if libRootDir=='RADIOHDL':
        print 'derive_all_use_libs = ', hdl.derive_lib_order(hdl.derive_all_use_libs('technology_memory'))
    if libRootDir=='UNB':
        print 'derive_all_use_libs = ', hdl.derive_lib_order(hdl.derive_all_use_libs('dp'))
        print 'derive_all_use_libs = ', hdl.derive_lib_order(hdl.derive_all_use_libs('uth'))
        print 'derive_all_use_libs = ', hdl.derive_lib_order(hdl.derive_all_use_libs('unb_common'))
        print 'derive_all_use_libs = ', hdl.derive_lib_order(hdl.derive_all_use_libs('unb_minimal'))

