###############################################################################
#
# Copyright (C) 2014
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

"""HDL configuration for building Modelsim simulation targets.
"""

import common as cm
import hdl_config
import sys
import os
import os.path

class ModelsimConfig(hdl_config.HdlConfig):

    def __init__(self, libRootDir, toolRootDir, libFileName='hdllib.cfg', toolFileName='hdltool.cfg'):
        """Get Modelsim tool info from toolRootDir and all HDL library info from libRootDir."""
        hdl_config.HdlConfig.__init__(self, libRootDir, toolRootDir, libFileName, toolFileName)
        # Modelsim

    def read_compile_order_from_mpf(self, mpfPathName):
        """Utility to read the compile order of the project files from an existing <mpfPathName>.mpf and save it into hdl_order.out."""
        # read <mpfPathName>.mpf to find all project files
        project_file_indices = []
        project_file_names = []
        with open(mpfPathName, 'r') as fp:
            for line in fp:
                words = line.split()
                if len(words)>0:
                    key = words[0]
                    if key.find('Project_File_')>=0 and key.find('Project_File_P_')==-1:
                        project_file_indices.append(key[len('Project_File_'):])
                        project_file_names.append(words[2])
        # read <mpfPathName>.mpf again to find compile order for the project files
        compile_order = range(len(project_file_names))
        with open(mpfPathName, 'r') as fp:
            for line in fp:
                words = line.split()
                if len(words)>0:
                    key = words[0]
                    if key.find('Project_File_P_')>=0:
                        project_file_index = project_file_indices.index(key[len('Project_File_P_'):])
                        project_file_name = project_file_names[project_file_index]
                        k = words.index('compile_order')
                        k = int(words[k+1])
                        compile_order[k]=project_file_name
        return compile_order

    def create_modelsim_project_file(self, technology_name, lib_names=None):
        """Create the Modelsim project file for all HDL libraries in the specified list of lib_names."""
        if lib_names==None: lib_names=self.lib_names
        lib_dicts = self.libs.get_dicts('hdl_lib_name', lib_names)
        for lib_dict in cm.listify(lib_dicts):
            # Open mpf
            lib_name = lib_dict['hdl_lib_name']
            mpf_name = lib_name + '.mpf'
            mpf_path = self.get_lib_build_dirs('sim', lib_dicts=lib_dict)
            cm.mkdir(mpf_path)
            mpfPathName = os.path.join(mpf_path, mpf_name)
            with open(mpfPathName, 'w') as fp:
                # Write [Library] section for all used libraries
                fp.write('[Library]\n')
                # . vendor technology libs
                tech_dict = self.read_modelsim_technology_libraries_file(technology_name)
                for lib_clause, lib_work in tech_dict.iteritems():
                    fp.write('%s = %s\n' % (lib_clause, lib_work))
                # . all used libs for this lib_name
                use_lib_names = self.derive_all_use_libs(lib_name)
                use_lib_dicts = self.libs.get_dicts('hdl_lib_name', use_lib_names)
                use_lib_build_sim_dirs = self.get_lib_build_dirs('sim', lib_dicts=use_lib_dicts)
                use_lib_clause_names = self.libs.get_key_values('hdl_library_clause_name', use_lib_dicts)
                for lib_clause, lib_dir in zip(cm.listify(use_lib_clause_names), cm.listify(use_lib_build_sim_dirs)):
                    lib_work = os.path.join(lib_dir, 'work')
                    fp.write('%s = %s\n' % (lib_clause, lib_work))
                # . work
                fp.write('work = work\n')
                # . others modelsim default libs
                model_tech_dir = os.path.expandvars(self.tool.get_key_values('model_tech_dir'))
                fp.write('others = %s\n' % os.path.join(model_tech_dir, 'modelsim.ini'))
                # Write [Project] section for all used libraries
                fp.write('[Project]\n')
                fp.write('Project_Version = 6\n')  # must be >= 6 to fit all
                fp.write('Project_DefaultLib = work\n')
                fp.write('Project_SortMethod = unused\n')
                # - hdl files
                synth_files = lib_dict['synth_files'].split()
                test_bench_files = lib_dict['test_bench_files'].split()
                sim_files = synth_files + test_bench_files
                fp.write('Project_Files_Count = %d\n' % len(sim_files))
                lib_path = self.libs.get_filePath(lib_dict)
                for i, fn in enumerate(sim_files):
                    fpn = os.path.expandvars(fn)                    # support using environment variables in the file path
                    if os.path.isabs(fpn):
                        filePathName = fpn                          # use absolute path to file
                    else:
                        filePathName = os.path.join(lib_path, fpn)  # derive absolute path to file from library path and local path to file
                    fp.write('Project_File_%d = %s\n' % (i, filePathName))
                project_file_p_defaults = 'vhdl_novitalcheck 0 file_type vhdl group_id 0 cover_nofec 0 vhdl_nodebug 0 vhdl_1164 1 vhdl_noload 0 vhdl_synth 0 vhdl_enable0In 0 vlog_1995compat 0 last_compile 0 vhdl_disableopt 0 cover_excludedefault 0 vhdl_vital 0 vhdl_warn1 1 vhdl_warn2 1 vhdl_explicit 0 vhdl_showsource 0 cover_covercells 0 vhdl_0InOptions {} vhdl_warn3 1 vlog_vopt {} cover_optlevel 3 voptflow 1 vhdl_options {} vhdl_warn4 1 toggle - ood 0 vhdl_warn5 1 cover_noshort 0 compile_to work cover_nosub 0 dont_compile 0 vhdl_use93 2002 cover_stmt 1'
                project_folders = []
                nof_synth_files = len(synth_files)
                if nof_synth_files>0:
                    project_folders.append('synth')
                    for i in range(nof_synth_files):
                        fp.write('Project_File_P_%d = folder %s compile_order %d %s\n' % (i, project_folders[-1], i, project_file_p_defaults))
                nof_test_bench_files = len(test_bench_files)
                if nof_test_bench_files>0:
                    project_folders.append('test_bench')
                    for j in range(nof_test_bench_files):
                        fp.write('Project_File_P_%d = folder %s compile_order %d %s\n' % (i+j+1, project_folders[-1], i+j+1, project_file_p_defaults))
                fp.write('Project_Folder_Count = %d\n' % len(project_folders))
                for i, fd in enumerate(project_folders):
                    fp.write('Project_Folder_%d = %s\n' % (i, fd))
                    fp.write('Project_Folder_P_%d = folder {Top Level}\n' % i)
                # - simulation configurations
                fp.write('Project_Sim_Count = %d\n' % len(test_bench_files))
                project_sim_p_defaults = 'Generics {} timing default -std_output {} -nopsl 0 +notimingchecks 0 -L {} selected_du {} -hazards 0 -sdf {} ok 1 -0in 0 -nosva 0 +pulse_r {} -absentisempty 0 OtherArgs {} -multisource_delay {} +pulse_e {} vopt_env 1 -coverage 0 -sdfnoerror 0 +plusarg {} -vital2.2b 0 -t default -memprof 0 is_vopt_flow 0 -noglitch 0 -nofileshare 0 -wlf {} -assertdebug 0 +no_pulse_msg 0 -0in_options {} -assertfile {} -sdfnowarn 0 -Lf {} -std_input {}'
                project_sim_p_optimization = 'is_vopt_opt_used 2'  # = when 'Enable optimization' is not selected in GUI
                project_sim_p_optimization = 'is_vopt_opt_used 1 voptargs {OtherVoptArgs {} timing default VoptOutFile {} -vopt_keep_delta 0 -0in 0 -fvopt {} VoptOptimize:method 1 -vopt_00 2 +vopt_notimingcheck 0 -Lfvopt {} VoptOptimize:list .vopt_opt.nb.canvas.notebook.cs.page1.cs.g.spec.listbox -Lvopt {} +vopt_acc {} VoptOptimize .vopt_opt.nb.canvas.notebook.cs.page1.cs -vopt_hazards 0 VoptOptimize:Buttons .vopt_opt.nb.canvas.notebook.cs.page1.cs.g.spec.bf 0InOptionsWgt .vopt_opt.nb.canvas.notebook.cs.page3.cs.zf.ze -0in_options {}}' # = when 'Enable optimization' is selected in GUI for full visibility
                for i, fn in enumerate(test_bench_files):
                    fName = os.path.basename(fn)
                    tbName = os.path.splitext(fName)[0]
                    fp.write('Project_Sim_%d = %s\n' % (i, tbName))
                for i, fn in enumerate(test_bench_files):
                    fName = os.path.basename(fn)
                    tbName = os.path.splitext(fName)[0]
                    fp.write('Project_Sim_P_%d = folder {Top Level} additional_dus work.%s %s %s\n' % (i, tbName, project_sim_p_defaults, project_sim_p_optimization))
                # Write [vsim] section
                fp.write('[vsim]\n')
                fp.write('RunLength = 0 ps\n')
                fp.write('resolution = 1ps\n')
                fp.write('IterationLimit = 100\n')
                fp.write('DefaultRadix = decimal\n')
                
    def read_modelsim_technology_libraries_file(self, technologyName, filePath=None):
        """Read the list of technology libraries from a file."""
        fileName = 'modelsim_libraries_' + technologyName + '.txt'             # use fixed file name format
        if filePath==None:
            fileNamePath=os.path.join(self.toolRootDir, 'modelsim', fileName)  # default file path
        else:
            fileNamePath=os.path.join(filePath, fileName)                      # specified file path
        tech_dict = self.tool.read_dict_file(fileNamePath)
        return tech_dict
    
    def create_modelsim_project_files_file(self, filePath=None, lib_names=None):
        """Create file with list of the Modelsim project files for all HDL libraries in the specified list of lib_names."""
        fileName = 'modelsim_project_files.txt'                                # use fixed file name
        if filePath==None:
            fileNamePath=os.path.join(self.toolRootDir, 'modelsim', fileName)  # default file path
        else:
            fileNamePath=os.path.join(filePath, fileName)                      # specified file path
        if lib_names==None: lib_names=self.lib_names
        with open(fileNamePath, 'w') as fp:
            lib_dicts = self.libs.get_dicts('hdl_lib_name', lib_names)
            mpf_paths = self.get_lib_build_dirs('sim', lib_dicts=lib_dicts)
            for lib_name, mpf_path in zip(cm.listify(lib_names),cm.listify(mpf_paths)):
                fp.write('%s = %s\n' % (lib_name, mpf_path))


if __name__ == '__main__':
    # Read the dictionary info from all HDL tool and library configuration files in the current directory and the sub directories
    libRootDir = 'RADIOHDL'
    #libRootDir = 'UNB'
    technology_name = 'stratixiv'
    msim = ModelsimConfig(libRootDir=os.environ[libRootDir], toolRootDir=os.path.expandvars('$RADIOHDL/tools'), libFileName='hdllib.cfg', toolFileName='hdltool.cfg')
    
    print '#'
    print '# ModelsimConfig:'
    print '#'
    print ''
    print 'HDL library paths that are found in $%s:' % libRootDir
    for p in msim.libs.filePaths:
        print '    ', p
        
    print ''
    print 'Derive library compile order = ', msim.derive_lib_order()

    print ''
    print 'get_lib_build_dirs for simulation:'
    for sim_dir in msim.get_lib_build_dirs('sim'):
        print '    ', sim_dir

    print ''
    print 'Create library compile order files.'
    msim.create_lib_order_files()
    
    print ''
    print 'Create modelsim projects list file.'
    msim.create_modelsim_project_files_file()
    
    print ''
    print 'Create modelsim project files for technology %s and all HDL libraries in $%s.' % (technology_name, libRootDir)
    msim.create_modelsim_project_file(technology_name)
    
    # Use save_compile_order_in_mpf = True to avoid having to manually edit the compile order in the hdllib.cfg, because
    # the synth_files need to be in hierarchical order. The test_bench_files are typically independent so these may be
    # put in alphabetical order. The hdllib.cfg still does need some manual editing to set the proper key and paths.
    lib_name = 'mm'
    save_compile_order_in_mpf = False 
    if libRootDir=='UNB' and save_compile_order_in_mpf:
        # Read compile order from existing <lib_name>.mpf
        mpfPathName = os.path.expandvars('$UNB/Firmware/designs/%s/build/synth/quartus/sopc_%s_sim/%s.mpf' % (lib_name, lib_name, lib_name))
        mpfPathName = os.path.expandvars('$UNB/Firmware/modules/Lofar/%s/build/sim/modelsim/%s.mpf' % (lib_name, lib_name))
        mpfPathName = os.path.expandvars('$UNB/Firmware/modules/%s/build/sim/modelsim/%s.mpf' % (lib_name, lib_name))
        compile_order = msim.read_compile_order_from_mpf(mpfPathName)
        # Append the compile_order list to the lib_name dictionary hdllib.cfg file
        lib_dict = msim.libs.get_dicts('hdl_lib_name', lib_name)
        lib_path = msim.libs.get_filePath(lib_dict)
        filePathName = os.path.join(lib_path, 'hdllib.cfg')
        print ''
        print 'Save modelsim compile order for', lib_name, 'in HDL library config file', filePathName
        msim.libs.append_key_to_dict_file(filePathName, 'files', compile_order)
